import logging

import numpy as np
import pandas as pd
from astropy import units as u
from astropy.time import Time
from tqdm import tqdm

from nuztf.observations.depot import coverage_depot_path, write_coverage_depot
from nuztf.observations.mns import MNS
from nuztf.observations.shared import coverage_dir, partial_flag
from nuztf.observations.skyvision import (
    coverage_skyvision_path,
    write_coverage_skyvision,
)
from nuztf.observations.tap import coverage_tap_path, write_coverage_tap

logger = logging.getLogger(__name__)


# The following functions are used to get the coverage from the cache


def get_coverage(jds: [int], backend="best") -> pd.DataFrame | None:
    """
    Get a dataframe of the coverage for a list of JDs

    Will use the cache if available, otherwise will query the depot, and lastly TAP

    :param jds: JDs
    :param backend: "best" or "depot" or "tap" or "skyvision" or "masterlog"
    :return: Coverage dataframe
    """

    assert backend in [
        "best",
        "depot",
        "tap",
        "skyvision",
        "masterlog",
    ], f"Invalid backend '{backend}'"

    # Clear any logs flagged as partial/incomplete

    cache_files = coverage_dir.glob("*.json")
    partial_logs = [x for x in cache_files if partial_flag in str(x)]

    if len(partial_logs) > 0:
        logger.debug(f"Removing the following partial logs: {partial_logs}")
        for partial_log in partial_logs:
            partial_log.unlink()

    # Only write missing logs

    covered_jds = []

    if backend in ["best", "depot"]:
        for jd in jds:
            if jd not in covered_jds:
                depot_path = coverage_depot_path(jd)
                if depot_path.exists():
                    df = pd.read_json(depot_path)
                    if len(df) > 0:
                        covered_jds.append(jd)

        missing_logs = sorted(set(jds) - set(covered_jds))

        if len(missing_logs) > 0:
            logger.info(
                f"Some logs were missing from the cache. "
                f"Querying for the following JDs in depot: {missing_logs}"
            )
            write_coverage_depot(missing_logs)

    if backend in ["best", "skyvision"]:
        for jd in jds:
            if jd not in covered_jds:
                skyvision_path = coverage_skyvision_path(jd)
                if skyvision_path.exists():
                    df = pd.read_json(coverage_skyvision_path(jd))
                    if len(df) > 0:
                        covered_jds.append(jd)

        missing_logs = sorted(set(jds) - set(covered_jds))

        if len(missing_logs) > 0:
            logger.info(
                f"Some logs were still missing from the cache. "
                f"Querying for the following JDs from skyvision: {missing_logs}"
            )
            write_coverage_skyvision(missing_logs)

    # Try TAP for missing logs

    if backend in ["best", "tap"]:
        for jd in jds:
            if jd not in covered_jds:
                tap_path = coverage_tap_path(jd)
                if tap_path.exists():
                    df = pd.read_json(tap_path)
                    if len(df) > 0:
                        covered_jds.append(jd)

        missing_logs = sorted(set(jds) - set(covered_jds))

        if len(missing_logs) > 0:
            logger.info(
                f"Some logs were still missing from the cache. "
                f"Querying for the following JDs from TAP: {missing_logs}"
            )
            write_coverage_tap(missing_logs)

    # Load logs from cache

    results = []

    for jd in tqdm(jds):

        res = None

        if backend in ["best", "depot"]:
            df = pd.read_json(coverage_depot_path(jd))
            if len(df) > 0:
                res = df

        if backend in ["best", "skyvision"]:
            if res is None:
                df = pd.read_json(coverage_skyvision_path(jd))
                if len(df) > 0:
                    res = df

        if backend in ["best", "tap"]:
            if res is None:
                df = pd.read_json(coverage_tap_path(jd))
                if len(df) > 0:
                    res = df

        if res is not None:
            results.append(res)

    if results:
        result_df = pd.concat(results, ignore_index=True)
        return result_df
    else:
        return None


def get_obs_summary_depot(t_min: Time, t_max: Time, backend="best") -> MNS | None:
    """
    Get observation summary from depot

    :param t_min: Start time
    :param t_max: End time
    :param backend: "best" or "depot" or "tap" or "skyvision"
    :return: MNS object
    """

    jds = np.arange(int(t_min.jd) - 0.5, int(t_max.jd) + 1.5)

    res = get_coverage(jds, backend=backend)

    if len(res) == 0:
        return None

    res["date"] = Time(res["obsjd"].to_numpy(), format="jd").isot

    mns = MNS(df=res)

    mns.data.query(f"obsjd >= {t_min.jd} and obsjd <= {t_max.jd}", inplace=True)

    mns.data.reset_index(inplace=True)
    mns.data.drop(columns=["index"], inplace=True)

    return mns


def get_obs_summary(
    t_min,
    t_max=None,
    max_days: float = None,
    backend="best",
) -> MNS | None:
    """
    Get observation summary from IPAC depot

    :param t_min: Start time
    :param t_max: End time
    :param max_days: Maximum number of days
    :param backend: "best" or "depot" or "tap" or "skyvision"
    :return: MNS object
    """
    now = Time.now()

    if t_max and max_days:
        raise ValueError("Choose either t_max or max_days, not both")

    if t_max is None:
        if max_days is None:
            t_max = now
        else:
            t_max = t_min + (max_days * u.day)

    if t_max > now:
        t_max = now

    logger.info(f"Getting observation logs  using backend {backend}.")
    mns = get_obs_summary_depot(t_min=t_min, t_max=t_max, backend=backend)

    if mns is not None:
        logger.info(f"Found {len(set(mns.data))} observations in total.")
    else:
        logger.warning(f"Found no observations using backend {backend}.")

    return mns



"""
Shared variables for the observations module
"""

from pathlib import Path

from astropy.time import Time
from ztfquery.io import LOCALSOURCE

coverage_dir = Path(LOCALSOURCE).joinpath("all_obs")
coverage_dir.mkdir(exist_ok=True, parents=True)

partial_flag = "_PARTIAL"


class NoDepotEntry(Exception):
    """
    No entry in the depot for a given date
    """


def get_date(jd: float) -> str:
    """
    Get the date in YYYYMMDD format from a JD

    :param jd: JD
    :return: String date
    """
    return str(Time(jd, format="jd").isot).split("T")[0].replace("-", "")



import json
import logging
from pathlib import Path

import requests
from astropy.time import Time
from requests.auth import HTTPBasicAuth
from tqdm import tqdm

from nuztf import credentials
from nuztf.observations.shared import NoDepotEntry, coverage_dir, get_date, partial_flag

logger = logging.getLogger(__name__)

username, password = credentials.load_credentials("ipacdepot")
data = {"username": username, "password": password}


def download_depot_log(date):
    """
    Download the depot log for a given date

    :param date: date in YYYYMMDD format
    :return: json log
    """
    url = f"https://ztfweb.ipac.caltech.edu/ztf/depot/{date}/ztf_recentproc_{date}.json"
    response = requests.get(url, auth=HTTPBasicAuth(username, password))
    if response.status_code == 404:
        raise NoDepotEntry(f"No depot entry for {date} at url {url}")
    response.raise_for_status()
    return response.json()


def coverage_depot_path(jd: float) -> Path:
    """
    Return the path to the cached coverage file for a given JD

    :param jd: JD
    :return: path to cached coverage file
    """

    date = get_date(jd)

    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    return coverage_dir.joinpath(f"{date}{partial_ext}.json")


def write_coverage_depot(jds: list[float]):
    """
    Write the depot coverage for a list of JDs to the cache

    Requires a valid IPAC depot login

    :param jds: JDs
    :return: None
    """
    for jd in tqdm(jds):
        date = get_date(jd)
        try:
            log = download_depot_log(date)
        except NoDepotEntry as exc:
            log = {}
            logger.warning(f"No depot entry for {date}: {exc}")

        path = coverage_depot_path(jd)
        with open(path, "w") as f:
            json.dump(log, f)



"""
Skyvision coverage
"""

import json
from pathlib import Path

import pandas as pd
from astropy.time import Time
from tqdm import tqdm
from ztfquery import skyvision

from nuztf.observations.shared import coverage_dir, get_date, partial_flag


def coverage_skyvision_path(jd: float) -> Path:
    """
    Find the path to the Skyvision coverage file for a given JD

    :param jd: JD
    :return: Output path
    """
    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    output_path = coverage_dir.joinpath(f"{get_date(jd)}{partial_ext}_skyvision.json")

    return output_path


def write_coverage_skyvision(jds: list[float]):
    """
    Write the Skyvision coverage for a list of JDs to the cache

    :param jds: JDs
    :return: None
    """
    for jd in tqdm(jds):
        date = get_date(jd)

        assert jd - int(jd) == 0.5, "JD must be a half-integer"

        res = skyvision.get_log(f"{date[:4]}-{date[4:6]}-{date[6:8]}", verbose=False)

        path = coverage_skyvision_path(jd)

        if res is not None:
            # The skyvision log has a bug where some entries have a FieldID of "NONE"
            mask = (
                pd.notnull(res["FieldID"]) & res["FieldID"].astype(str).str.isnumeric()
            )
            res = res[mask]
            jds = [
                Time(f'{row["UT Date"]}T{row["UT Time"]}').jd
                for _, row in res.iterrows()
            ]
            res["obsjd"] = jds
            res["status"] = (res["Observation Status"] == "FAILED").astype(int)
            res["filter_id"] = res["Filter"].apply(
                lambda x: 1 if x == "FILTER_ZTF_G" else 2 if x == "FILTER_ZTF_R" else 3
            )
            res["maglim"] = 20.5
            res["field_id"] = res["FieldID"].astype(int)
            res["exposure_time"] = res["Exptime"]
            res = res[
                ["obsjd", "filter_id", "field_id", "exposure_time", "maglim", "status"]
            ]

            new_res = []
            for _, row in res.iterrows():
                new = row.to_dict()
                new_res += [dict(qid=int(i), **new) for i in range(64)]
            pd.DataFrame(new_res).to_json(path)

        else:
            with open(path, "w") as f:
                json.dump({}, f)



"""
Module for handling the master log of observations.
"""

import pandas as pd
import requests
from astropy.time import Time
from requests.auth import HTTPBasicAuth

from nuztf.credentials import load_credentials
from nuztf.paths import ZTF_LOG_PATH

MASTER_LOG_URL = "https://sites.astro.caltech.edu/~tb/ztfops/sky/allexp.tbl"

username, password = load_credentials("ztfops")


def download_master_log():
    """
    Download the master log of observations
    """
    response = requests.get(MASTER_LOG_URL, auth=HTTPBasicAuth(username, password))
    response.raise_for_status()
    with open(ZTF_LOG_PATH, "wb") as f:
        f.write(response.content)


def load_master_log() -> pd.DataFrame:
    """
    Load the master log of observations
    """
    df = pd.read_fwf(ZTF_LOG_PATH, comment="|")
    mask = df["type"] == "targ"
    df = df[mask]
    mask = df["p"].astype(float) > 0
    df = df[mask]
    df.rename(
        columns={"l": "filter_id", "ld": "field_id", "p": "exposure_time", "id": "pid"},
        inplace=True,
    )
    obs = Time(df["UT_START"].to_list(), format="isot")
    df["obsjd"] = obs.jd
    df["status"] = 0
    df["maglim"] = 20.5

    df = df[["obsjd", "filter_id", "field_id", "exposure_time", "maglim", "status"]]
    df.reset_index(inplace=True, drop=True)
    return df



import logging
import time
from pathlib import Path

import backoff
import numpy as np
import pyvo.dal
import requests
from astropy.time import Time
from pyvo.auth import authsession, securitymethods
from pyvo.utils.http import create_session

from nuztf.credentials import load_credentials
from nuztf.observations.shared import coverage_dir, get_date, partial_flag

logger = logging.getLogger(__name__)


# IPAC TAP login
username, password = load_credentials("irsa")
data = {"username": username, "password": password}
headers = {"Content-Type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
IPAC_TAP_URL = "https://irsa.ipac.caltech.edu/TAP"

# The following functions are used to find the paths to the cached coverage files


def coverage_tap_path(jd: float) -> Path:
    """
    Find the path to the TAP coverage file for a given JD

    :param jd: JD
    :return: Output path
    """
    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    output_path = coverage_dir.joinpath(f"{get_date(jd)}{partial_ext}_tap.json")

    return output_path


# The following functions are used to write the coverage to the cache


@backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=60)
def write_coverage_tap(jds: [float]):
    """
    Write the coverage for a list of JDs to the cache using TAP

    (JDs must be half-integer)

    :param jds: Half-integer JDs
    :return: None
    """
    with create_session() as session:
        session.auth = (username, password)

        # Setup AuthSession
        auth = authsession.AuthSession()
        auth.credentials.set(securitymethods.BASIC, session)
        auth.add_security_method_for_url(IPAC_TAP_URL, securitymethods.BASIC)
        auth.add_security_method_for_url(IPAC_TAP_URL + "/sync", securitymethods.BASIC)
        auth.add_security_method_for_url(IPAC_TAP_URL + "/async", securitymethods.BASIC)
        auth.add_security_method_for_url(
            IPAC_TAP_URL + "/tables", securitymethods.BASIC
        )

        # Create service
        client = pyvo.dal.TAPService(
            "https://irsa.ipac.caltech.edu/TAP", session=session
        )

        for jd in jds:
            assert jd - int(jd) == 0.5, "JD must be a half-integer"

            obstable = client.search(
                f"""
            SELECT expid,obsjd,fid,field,exptime,rcid,maglimit,infobits,ipac_gid
            FROM ztf.ztf_current_meta_sci WHERE (obsjd BETWEEN {jd} AND {jd + 1})
            """
            ).to_table()
            names = (
                "expid",
                "exptime",
                "field",
                "fid",
                "rcid",
                "maglimit",
                "infobits",
                "ipac_gid",
            )
            renames = (
                "exposure_id",
                "exposure_time",
                "field_id",
                "filter_id",
                "qid",
                "maglim",
                "status",
                "programid",
            )
            obstable.rename_columns(names, renames)

            obs = obstable.to_pandas()
            obs["nalertpackets"] = np.nan

            output_path = coverage_tap_path(jd)

            obs.to_json(output_path)
            time.sleep(1)

            logger.info(f"Coverage for JD {jd} written to {output_path} using TAP.")
